/* Time-stamp: <2005-01-05 13:22:56 poser> */

/*
 * Copyright (C) 1993-2003 William J. Poser.
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

static char rcsid[] = "$Id: misc.c,v 1.10 2003/11/20 02:51:49 poser Exp $";

#include <stdio.h>
#include <string.h>
#include "exitcode.h"
#include "retcodes.h"
#include "ex_codes.h"
#include "retcodes.h"
#include "defs.h"

void
DumpExclusions(unsigned char *exct,int exctsize,FILE *fp)
{
  int i;
  unsigned char flags;
	
  fprintf(fp,"  Characters excluded:\n");
  if(exctsize > 1) printf("\n");
  for(i = 0;i < exctsize; i++){
    if(exct[i]){
      flags = exct[i];
      fprintf(fp,"\t\t   %0.3o",i);
      if(flags & EX_INITIAL) fprintf(fp,"\tinitial");
      if(flags & EX_MEDIAL) fprintf(fp,"\tmedial");
      if(flags & EX_FINAL) fprintf(fp,"\tfinal");
      fprintf(fp,"\n");
    }
  }
}


/*
 * Overwrite a textual key with itself, skipping characters specified
 * in the exclusion table.
 */

int
Exclude(KEYTYPE *str, int len, unsigned char *exct)
{
  int i;
  int cnt;

  cnt = 0;
  if((str[0] >= ASCIISIZE) || (!(exct[str[0]] & EX_INITIAL)) ){
    str[cnt++] = str[0];
  }
  for(i = 1; i < len-1; i++){
    if( (str[i] < ASCIISIZE) && (exct[str[i]] & EX_MEDIAL)) continue;
    else str[cnt++] = str[i];
  }
  if((str[len-1] >= ASCIISIZE) || (!(exct[str[len-1]] & EX_FINAL)) ){
    str[cnt++] = str[len-1];
  }
  str[cnt] = ((KEYTYPE) 0); /* Null terminate */
  return(cnt);
}

void
KeyCopy(KEYTYPE *t,KEYTYPE *s,int len)
{
  register int i;
  for(i = 0;i < len;i++) *t++ = *s++;
}

void
RevKeyCopy(KEYTYPE *t,KEYTYPE *s,int len)
{
  register int i;
	
  s += (len-2);
	
  for(i = 0;i < len-1;i++) *t++ = *s--;
  *t= (KEYTYPE) 0;
}

void
CopyCommandLine(FILE *fp,int ac, char **av)
{
  int i;
	
  for(i=0; i < ac-1; i++){
    fprintf(fp,"%s ",av[i]);
  }
  fprintf(fp,"%s\n",av[ac-1]);
}

/* Given a date string, write the equivalent double into key. */

int
GetDateKey(char *field, double *key, struct ymdinfo *ymd)
{
  int num[3];
  double month;
  double day;
	
	
  if(sscanf(field,ymd->fmt,&(num[0]),&(num[1]),&(num[2])) != 3)
    return(ERROR);
  month = (double) num[ymd->m];
  day = (double) num[ymd->d];
  if(month < 0 || day < 0) return (ERROR);
  if(month > 12) return (ERROR);
  if(day > 31) return (ERROR);
  *key = (double) ((373 * (double) num[ymd->y]) + (31 * month) + day);
  return(SUCCESS);
}


/* Given a time string, write the equivalent double into key. */

int
GetTimeKey(char *field, double *key)
{
  int hours;
  int minutes;
  int seconds;
	
  seconds = 0;
  if(field[strlen(field)-3] == '.'){
    if(sscanf(field,"%d:%d.%d",&hours,&minutes,&seconds) != 3)
      return(ERROR);
  }
  else{
    if(sscanf(field,"%d:%d",&hours,&minutes) != 2)
      return(ERROR);
  }

  if(hours < 0 || minutes < 0 || seconds < 0) return (ERROR);
  if(hours > 24) return(ERROR);
  if(minutes > 60 || seconds > 60) return (ERROR);
	
  *key = (double) (((((double) hours * 60) + (double) minutes) * 60) + (double) seconds);
  return(SUCCESS);
}

void
IdentifySelf(FILE *fp)   
{
  extern char progname[];
  extern char version[];
  extern char* compdate;
	
  fprintf(fp,"%s %s\n",progname,version);
  fprintf(fp,"Compiled %s\n",compdate);
}
    
FILE *
OpenFile(char *file,char *mode,char *progname)
{
  FILE *fp;
  extern FILE *fopen();
  static char modestr[7];
   
  if((fp = fopen(file,mode)) != NULL) return fp;
  else{
    if(*mode == 'r') sprintf(modestr,"read");
    else if(*mode == 'w') sprintf(modestr,"write");
    else sprintf(modestr,"append");
    fprintf(stderr,"%s: can't open file %s for %s.\n",progname,file,modestr);
    exit(OPENERROR);
  }
  /* NOTREACHED */
}

/*
 * Rewrite a string, replacing backslash escapes.
 *
 */

#define DEFAULT        0
#define READBACKSLASH  1
#define READONEDIGIT   2
#define READTWODIGITS  3

int
EvalEscapes(char *s)
{
  char c;
  char odigit1;
  char odigit2;
  int state;
  char sum;
  char *SStart;
  char *TStart;
  char *t;


  /* Allocate storage for de-escaped copy of string */
	
  t = (char *) malloc(strlen(s)+1);
  if(t == (char *) 0) return(ERROR);

  SStart = s;
  TStart = t;

  /* Work through string */
	
  state = 0;
  while((c = *s++) != '\0'){
    switch(c){
    case '\\':
      if(state == READBACKSLASH){
	*t++ = '\\'; /* Literal backslash */
	state = DEFAULT;
      }
      else if(state  == READONEDIGIT){
	*t++ = '\\';
	*t++ = odigit1;
	state = DEFAULT;
      }
      else if(state  == READTWODIGITS){
	*t++ = '\\';
	*t++ = odigit1;
	*t++ = odigit2;
	state = DEFAULT;
      }
      else{				/* state == DEFAULT */
	state = READBACKSLASH;
      }
      break;
    case 'n':
      if(state == READBACKSLASH){
	*t++ = '\n';
	state = DEFAULT;
      }
      else{
	if(state  == READONEDIGIT){
	  *t++ = '\\';
	  *t++ = odigit1;
	}
	else if(state  == READTWODIGITS){
	  *t++ = '\\';
	  *t++ = odigit1;
	  *t++ = odigit2;
	}
	*t++ = c;
	state = DEFAULT;
      }
      break;
    case ' ':
      if(state == READBACKSLASH){
	*t++ = ' ';
	state = DEFAULT;
      }
      else{
	if(state  == READONEDIGIT){
	  *t++ = '\\';
	  *t++ = odigit1;
	}
	else if(state  == READTWODIGITS){
	  *t++ = '\\';
	  *t++ = odigit1;
	  *t++ = odigit2;
	}
	*t++ = c;
	state = DEFAULT;
      }
      break;
    case 't':
      if(state == READBACKSLASH){
	*t++ = '\t';
	state = DEFAULT;
      }
      else{
	if(state  == READONEDIGIT){
	  *t++ = '\\';
	  *t++ = odigit1;
	}
	else if(state  == READTWODIGITS){
	  *t++ = '\\';
	  *t++ = odigit1;
	  *t++ = odigit2;
	}
	*t++ = c;
	state = DEFAULT;
      }
      break;
    case '0':
    case '1':
    case '2':
    case '3':
      if(state == READBACKSLASH){
	odigit1 = c;
	sum = ( (c - '0') * 64);
	state = READONEDIGIT;
      }
      else if(state == READONEDIGIT){
	odigit2 = c;
	sum+=( (c - '0') * 8);
	state = READTWODIGITS;
      }
      else if(state == READTWODIGITS){
	sum+= (c - '0');
	*t++ = sum;
	state = DEFAULT;
      }
      else *t++ = c;
      break;
    case '4':
    case '5':
    case '6':
    case '7':
      /* These are not possible leading digits */
      if(state == READBACKSLASH){
	*t++ = '\\';
	*t++ = c;
	state = DEFAULT;
      }
      else if(state == READONEDIGIT){
	odigit2 = c;
	sum+=( (c - '0') * 8);
	state = READTWODIGITS;
      }
      else if(state == READTWODIGITS){
	sum+= (c - '0');
	*t++ = sum;
	state = DEFAULT;
      }
      else *t++ = c;
      break;
    default:
      if(state == READBACKSLASH){
	*t++ = '\\';
      }
      else if(state == READONEDIGIT){
	*t++ = '\\';
	*t++ = odigit1;
      }
      else if(state == READTWODIGITS){
	*t++ = '\\';
	*t++ = odigit1;
	*t++ = odigit2;
      }
      *t++ = c;
      state = DEFAULT;
      break;
    }
  }
  
  /* Null terminate copy and overwrite original */
  
  *t = '\0';
  strcpy(SStart,TStart);
  free((void *) TStart);
  return(SUCCESS);
}

/*
 * Given a number represented as a string, insert delimiters to break it up for
 * readability. Normally, the delimiter will be a comma which will be inserted every
 * three digits. However, the delimiter and groupsize are arguments, permitting use
 * in other locales.
 * 
 * The string is assumed to consist of digits, possibly preceded by spaces,
 * and possibly containing a decimal point, i.e.: [:space:]*[:digit:]*\.[:digit:]*
 * 
 * The delimited string is returned in freshly allocated memory, or NULL storage
 * cannot be allocated. 
 */

char * DelimitNumber(char *number, char delim, int groupsize)
{
  int newsize;			/* Characters needed for delimited string */
  int length;			/* Length of input string */
  int commas;			/* Number of delimiters to insert */
  int commacnt;			/* Number of delimiters inserted so far */
  int digits;			/* Number of digits preceding decimal point */
  int cnt;			/* Number of digits processed */
  char *new;			/* String in which to write delimited number */
  char *point;			/* Location of decimal point */
  char *pos;			/* Current position in original number */
  char *npos;			/* Current position in new number */
  char *NonSpaceNumber;		/* Location of first non-space character in number*/

  length = strlen(number);
  point = strrchr(number,'.');
  if(point == NULL) point = number+(length-1);
  else point = point -1;	/* Point now points at last digit preceding decimal */
  NonSpaceNumber = strrchr(number,040);
  if(NonSpaceNumber == NULL) NonSpaceNumber = number;
  else NonSpaceNumber+=1;
  digits = 1+point - NonSpaceNumber;
  commas = (digits-1)/groupsize;

  newsize = length + commas +1;	/* Add one for final null */
  new = (char *) malloc(newsize * sizeof(char));
  if(new){
    /* First copy right hand part of number, up to and including decimal point */
    pos = number+length+1;
    npos = new+newsize;
    for(; pos > point;) *npos-- = *pos--;

    /* Now copy lefthand part, inserting delimiters as we go */
    for(cnt=commacnt=0; pos >= NonSpaceNumber;){
      *npos-- = *pos--;
      cnt++;
      if( (cnt % groupsize == 0) && (commacnt++ < commas )) *npos-- = delim;
    }
    while(pos >= number) *npos-- = *pos--; /* Copy leading spaces, if any */
  }
  return new;
}
