/* Time-stamp: <2005-02-03 20:43:06 poser> */
static char rcsid[] = "$Id: msort.c,v 6.28 2004/10/27 04:17:27 poser Exp poser $";

/*
 * Copyright (C) 1993-2004 William J. Poser.
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * or go to the web page:  http://www.gnu.org/licenses/gpl.txt.
 *
 *
 * This program sorts text using an arbitrary sort order specified
 * in a file. Records may be either single lines or blocks of text.
 * Sorting may be on the entire record or on a specified field,.
 * where the set of field terminator characters may be specified
 * on the command line. The fields on which to sort may be specified
 * either by number or by the tags with which they begin.
 * Keys may be optional or obligatory.
 *
 * Author:  Bill Poser
 *	 
 */

#define PROGNAME "msort"
#define LOGNAME PROGNAME ".log"

#ifdef SEVENBIT
#define VBITS " (7 bit) "
#else
#define VBITS " (8 bit) "
#endif
char progname[]=PROGNAME;
char version[]="6.18";
char *compdate= __DATE__ " " __TIME__;

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#include <sys/types.h>
#include <sys/file.h>
#include "regexp.h"
#include "exitcode.h"
#include "retcodes.h"
#include "input.h"
#include "limits.h"
#include "defs.h"
#include "record.h"
#include "ex_codes.h"

#define REPORTINTERVAL 500
#define QUICKSORTTHRESHOLD 1000L
#define INSERTIONSORTTHRESHOLD 7L
#define INTWREPORTINTERVAL (4 * REPORTINTERVAL)

#define PRTEMPSTRLEN 32

/* How are exclusions specified? */

#define FILESPEC 1
#define STRINGSPEC 2

/* Comparison information bits */

#define CINVERSE (unsigned char) 0x01
#define CNUMERIC (unsigned char) 0x02
#define CDATE    (unsigned char) 0x04
#define CTIME    (unsigned char) 0x08
#define CSIZE    (unsigned char) 0x10

/*
 *  Missing key comparison options
 *  DO NOT CHANGE THESE VALUES! The code in Compare()
 *  depends on these specific values.
 */

#define COMPARE_LESSTHAN 0
#define COMPARE_EQUAL 1
#define COMPARE_GREATERTHAN 2

/* Size of rank table */
#define RANKTABLESIZE (MAXMULTIGRAPHS + ASCIISIZE)

/* Maximum index into rank table */
#define MAXCODE (RANKTABLESIZE)

#define TRUE	1
#define FALSE	0

/* Types of key field selections */

#define WHOLE 0
#define TAGGED 1
#define NUMBERED 2

/* Entry in the map table, for mapping sequences to codes */

struct mapentry{
   char *str;
   KEYTYPE code;
   size_t len;			/* Length of str */
};

/*
 * Each key is represented by one keyinfo structure.
 */

struct keyinfo{
  int SelectType;		/* How to select key field */
  regexp *cregexp;		/* Pointer to compiled tag regexp */
  short Number;		 	/* Which field to sort on if numbered */
  int *RankTable;
  unsigned char *ExclusionTable;
  int ExclusionSpec;		/* Have exclusions been specified on command line or in file? */
  struct mapentry **MapTable;
  int MapTableEntries;
  KEYTYPE NextCode;		/* Next code available in rank table */
  short OptionalP;		/* Is this key optional? */
  short MissingKeyComparison;	/* How missing keys compare to keys that are present */
  short ReverseP;		/* Reverse field? */
  unsigned char CompType;	/* Type of comparison to use */
  char *Tag;		        /* Tag of field to sort on if tagged */
  struct ymdinfo ymd;		/* Date format */
};

/* This is the array containing information about all the keys */

struct keyinfo KeyInfo[MAXKEYS];

/* This is the list of all of the records */

struct record **RecordList;

int VerboseP;		        /* Chat while we work */
int KeyCount;		/* Number of keys to sort on */
char *InputText;		/* Temporary storage for input text*/
KEYTYPE *TempString;	/* Temporary storage for text keys*/

char *Terminators;		/* Characters terminating fields */
int TerminatorCnt; /* Number of terminator characters */
int InitialMaxFields;	/* Initial value of MaxFields */

int ReverseP;		/* Invert the sense of comparisons */	
FILE *Logfp;			/* Log file */

long ComparisonCnt;		/* Calls to Compare */

main(int ac, char **av)
{

   int i;			/* Loop index */
   int InitialMaxRecords;	/* Initial value of MaxRecords */
   int MaxRecords;		/* Maximum records allocated */
   int InputTextSize;		/* Size of input text buffer */
   int TempStringSize;		/* Size of temp string buffer */
   long Records;		/* Number of records so far */
   int RecordLength;		/* Length of text in current record */
   int WholeKeyP;		/* Use whole record as key? */
   int opt;			/* Option returned by getopt() */
   int TerminatorsSetP;		/* Did user override default terminators? */
   char RecordSeparator;
   char *InputCopy;		/* Copy for reporting ill-formed keys on ext sorts */
   int tmpint;			/* Temporary integer */
   short FailedKeyP;		/* True if key extraction failed on some record */

   char *TagTemp = NULL;	/* Temporary string storage */
   unsigned int TagTempStringLength =0;	/* Current length of above */
   unsigned int TagTempLengthNeeded;
   char PrTempString[PRTEMPSTRLEN+1];

   char *tmpptr;		/* Temporary pointer */

   char *infile;		/* Name of input file */
   FILE *infp;			/* Input file */
   FILE *outfp;			/* Output file */

   char * (*GetRecord)(FILE *, char *, int *, int *, char);
	
   extern FILE *OpenFile(char *,char *,char *);
   extern char *copy_string(char *);
   extern char *create_string(int);
   extern int EvalEscapes(char *);
   int GetExclusions(char *,struct keyinfo *);
   int GetKeys(struct record *, struct keyinfo *,int);
   int GetSortOrder(char *,struct keyinfo *);
   int SetDateFormat(char *, struct ymdinfo *);
   void IntWrite(long, FILE *,int,char);
   void DescribeKeys(FILE *);
   void InitializeKeyInfo(void);
   void CreateRankTable(struct keyinfo *);
   void ProvideDefaultRanks(void);
   void Copyright(FILE *);
  
   extern int optind;		/* Set by getopt */
   extern char *optarg;		/* Set by getopt */

   void QuickSort(struct record **,long,long);
   void ShellSort(struct record **,long);
		
   extern char * GetNNBlockRA(FILE *, char *, int *, int *, char);
   extern char * GetBlockSepCharRA(FILE *, char *, int *, int *, char);

   extern void CopyCommandLine(FILE *, int, char **);
   extern void IdentifySelf(FILE *);
   extern void PrintDefaults(FILE *);
   extern void PrintFlags(FILE *);
   extern void PrintLimits(FILE *);
   extern void PrintUsage(void);
   extern char * DelimitNumber(char *,int, int);
   
   extern long ftell();
   extern int getopt();
   
   /* Executable statements begin here */
   
   if(!((ac > 1) && (strcmp(av[1],"-q") == 0))){
      IdentifySelf(stderr);
      Copyright(stderr);
   }	
   if(ac < 2){
      PrintUsage();
      exit(BADOPTION);
   }
	

   /* 
    * Initializations and defaults.
    *
    * Note that the defaults for field terminators are set
    * after the command line is processed since they depend
    * on whether records are lines or blocks.
    */


   Records = 0L;
   KeyCount = 0;
   ComparisonCnt = 0L;
   
   InitialMaxRecords = DEFMAXRECORDS;
   InitialMaxFields = DEFMAXFIELDS;
   InputTextSize = INPUTTEXTSIZEINIT;
   
   GetRecord = GetNNBlockRA;
   
   VerboseP = TRUE;
   ReverseP = FALSE;
   TerminatorsSetP = FALSE;
   WholeKeyP = FALSE;
   RecordSeparator = '\n';
   FailedKeyP = FALSE;
   
   InitializeKeyInfo();
   
   /* Allocate storage */
   
   InputText = (char *) malloc((unsigned) InputTextSize * sizeof(char));
   if(InputText == (char *) 0){
      fprintf(stderr,"%s: out of memory - InputText .\n",progname);
      exit(OUTOFMEMORY);
   }
   
   TempStringSize = InputTextSize;
   TempString = (KEYTYPE *) malloc((unsigned) TempStringSize * sizeof(KEYTYPE));
   if(TempString == (KEYTYPE *) 0){
      fprintf(stderr,"%s: out of memory - TempString.\n",progname);
      exit(OUTOFMEMORY);
   }
   
   /* Open log file */

   Logfp = OpenFile(LOGNAME,"w",progname);
   IdentifySelf(Logfp);
   Copyright(Logfp);

   fprintf(Logfp,"Invoked by command line:\n  ");
   CopyCommandLine(Logfp,ac,av);
   
   /* Process command line */
   
   while( (opt = getopt(ac,av,"bc:d:Df:FhiIlLM:n:o:qr:Rs:t:vwx:X:")) != EOF){
      switch(opt){
	 /* How records are delimited */
       case 'b':		/* A record is a newline-terminated block */
	 GetRecord = GetNNBlockRA;
	 break;
       case 'l':
	 GetRecord = GetBlockSepCharRA;
	 break;
       case 'r':		/* Specify record separator */
	 GetRecord = GetBlockSepCharRA;
	 if(EvalEscapes(optarg) == ERROR){
	    fprintf(stderr,"EvalEscapes: out of memory.\n");
	    exit(OUTOFMEMORY);
	 }
	 RecordSeparator = optarg[0];
	 break;
	 
	 /* How fields are delimited */
       case 'd':		/* Fields are terminated by the named characters */
	 if(EvalEscapes(optarg) == ERROR){
	    fprintf(stderr,"EvalEscapes: out of memory.\n");
	    exit(OUTOFMEMORY);
	 }
	 Terminators = copy_string(optarg);
	 TerminatorsSetP = TRUE;
	 break;
	 
	 /* What to sort on */
       case 't':		/* Next key has specified tag */
	 if(WholeKeyP){
	    fprintf(stderr,"Use of key field and whole record is inconsistent.\n");
	    exit(BADOPTION);
	 }
	 if(KeyCount >= MAXKEYS -1){
	    fprintf(stderr,"Limit of %d keys exceeded.\n",MAXKEYS);
	    exit(LIMITEXCEEDED);
	 }

	 KeyInfo[KeyCount].SelectType = TAGGED;

	 TagTempLengthNeeded = strlen(optarg)+strlen("()(.*)")+1;
	 if( TagTempLengthNeeded > TagTempStringLength){
	   if(TagTemp != NULL) free(TagTemp);
	   TagTemp = (char *) malloc((size_t) TagTempLengthNeeded);
	   if(TagTemp == NULL){
	     fprintf(stderr,"Cannot allocate storage for TagTemp string.\n");
	     exit(OUTOFMEMORY);
	   }
	 }	   
	 sprintf(TagTemp,"(%s)(.*)",optarg);
	 if(EvalEscapes(TagTemp) == ERROR){
	    fprintf(stderr,"EvalEscapes: out of memory.\n");
	    exit(OUTOFMEMORY);
	 }
	 KeyInfo[KeyCount].cregexp = regcomp(TagTemp);
	 if(KeyInfo[KeyCount].cregexp == NULL){
	    fprintf(stderr,"%s is an ill-formed regular expression.\n",
		    optarg);
	    fprintf(Logfp,"%s is an ill-formed regular expression.\n",
		    optarg);
	    exit(BADOPTIONARG);
	 }
	 KeyInfo[KeyCount].Tag = (char *) malloc(strlen(optarg)+1);
	 if(KeyInfo[KeyCount].Tag) strcpy(KeyInfo[KeyCount].Tag,optarg); /* For report */
	 else{
	    fprintf(stderr,"Out of memory.\n");
	    exit(OUTOFMEMORY);
	 }
	 ++KeyCount;
	 break;
       case 'n':		/* Next key is Nth field */
	 if(WholeKeyP){
	    fprintf(stderr,"Use of key field and whole record is inconsistent.\n");
	    exit(BADOPTION);
	 }
	 if(KeyCount >= MAXKEYS -1){
	    fprintf(stderr,"Limit of %d keys exceeded.\n",MAXKEYS);
	    exit(LIMITEXCEEDED);
	 }
	 tmpint = atoi(optarg);
	 if(tmpint == 0){
	   fprintf(stderr,"%s: 0 is not a possible field number\n",progname);
	   exit(BADOPTIONARG);
	 }
	 KeyInfo[KeyCount].SelectType = NUMBERED;
	 KeyInfo[KeyCount].Number = tmpint -1;
	 ++KeyCount;
	 break;
       case 'w':		/* Sort on the entire input text */
	 WholeKeyP=TRUE;
	 if(KeyCount > 0){
	    fprintf(stderr,"Use of key field and whole record is inconsistent.\n");
	    exit(BADOPTION);
	 }
	 KeyCount = 1;
	 KeyInfo[0].SelectType = WHOLE;
	 break;
       case 'i':
	 if(KeyCount <1){
	    fprintf(stderr,"Key inversion specified without previous key selector.\n");
	    exit(BADOPTION);
	 }
	 KeyInfo[KeyCount-1].CompType |= CINVERSE;
	 break;

       case 'R':
	 if(KeyCount <1){
	    fprintf(stderr,"Key reversal specified without previous key selector.\n");
	    exit(BADOPTION);
	 }
	 KeyInfo[KeyCount-1].ReverseP = TRUE;
	 break;

       case 'f':
	 if(KeyCount <1){
	    fprintf(stderr,"Date format specified without previous key selector.\n");
	    exit(BADOPTION);
	 }
	 if(SetDateFormat(optarg,&(KeyInfo[KeyCount-1].ymd)) == ERROR){
	    fprintf(stderr,"%s is not a valid date format.\n",optarg);
	    exit(BADOPTIONARG);
	 }
	 break;


       case 'o':
	 if(KeyCount <1){
	    fprintf(stderr,"Optionality specified without previous key selector.\n");
	    exit(BADOPTION);
	 }
	 else{
	   if(KeyInfo[0].SelectType == WHOLE){
	    fprintf(stderr,"If the whole record is the key, the key may not be optional.\n");
	    exit(BADOPTION);
	   }
	 }
	 KeyInfo[KeyCount-1].OptionalP = TRUE;
	 switch(optarg[0]){
	   case 'l':
  	   case 'L':
	   case '<':
	     KeyInfo[KeyCount-1].MissingKeyComparison = COMPARE_LESSTHAN;
	     break;
	   case 'g':
  	   case 'G':
	   case '>':
	     KeyInfo[KeyCount-1].MissingKeyComparison = COMPARE_GREATERTHAN;
	     break;
	   case 'e':
  	   case 'E':
	   case '=':
	   default:
	     KeyInfo[KeyCount-1].MissingKeyComparison = COMPARE_EQUAL;
	     break;
	 }
	 break;

	 /* What files to use */	
	 
       case 's':		/* Read the sort order from the named file */
	 if(KeyCount <1){
	    fprintf(stderr,"Sort type specified without previous key selector.\n");
	    exit(BADOPTION);
	 }
	 if(GetSortOrder(optarg,&(KeyInfo[KeyCount-1])) == ERROR){
					
	    fprintf(stderr,
		    "Attempt to read sort order from file %s failed.\n",optarg);
	    exit(OTHERERROR);
	 }
	 fprintf(Logfp,"Read sort order for key %d from file %s.\n",
		 KeyCount,optarg);
	 break;
				
       case 'x':		/* Read exclusions from the named file */
	 if(KeyInfo[KeyCount-1].ExclusionSpec == STRINGSPEC){
	   fprintf(stderr,"Exclusions already specified on command line for this key.\n");
	   exit(BADOPTION);
	 }
	 if(KeyCount <1){
	    fprintf(stderr,"Exclusions specified without previous key selector.\n");
	    exit(BADOPTION);
	 }
	 if(KeyInfo[KeyCount-1].CompType & CNUMERIC){
	    fprintf(stderr,"Exclusions don't make sense with a non-textual sort.\n");
	    exit(BADOPTION);
	 }
	 if(GetExclusions(optarg,&(KeyInfo[KeyCount-1])) != SUCCESS){
	    fprintf(stderr,
		    "Attempt to read exclusions from file %s failed.\n",optarg);
	    exit(OTHERERROR);
	 }
	 KeyInfo[KeyCount-1].ExclusionSpec=FILESPEC;
	 break;

       case 'X':		/* Read exclusions from command line */
	 if(KeyInfo[KeyCount-1].ExclusionSpec == FILESPEC){
	   fprintf(stderr,"Exclusions already specified in file for this key.\n");
	   exit(BADOPTION);
	 }
	 if(KeyCount <1){
	    fprintf(stderr,"Exclusions specified without previous key selector.\n");
	    exit(BADOPTION);
	 }
	 if(KeyInfo[KeyCount-1].CompType & CNUMERIC){
	    fprintf(stderr,"Exclusions don't make sense with a non-textual sort.\n");
	    exit(BADOPTION);
	 }
	 GetExclusionsString(optarg,&(KeyInfo[KeyCount-1]));	
	 KeyInfo[KeyCount-1].ExclusionSpec=STRINGSPEC;
	 break;
	 
	 /* Sizes of things */
	 
       case 'M':
	 InitialMaxRecords = atoi(optarg);
	 if(InitialMaxRecords <= 2){
	    fprintf(stderr,
		    "%s: invalid initial maximum record specification %s.\n",
		    progname,optarg);
	    exit(BADOPTIONARG);
	 }
	 break;
	 
	 /* Miscellaneous */
	 
       case 'I':		/* Reverse sense of comparisons globally */
	 ReverseP = TRUE;
	 break;
       case 'c':		/* Comparison type */
	 if(KeyCount <1){
					
	    fprintf(stderr,"Comparison type specified without previous key selector.\n");
	    exit(BADOPTION);
	 }
	 switch(optarg[0]){
	  case 'd':
	  case 'D':
	    KeyInfo[KeyCount-1].CompType |= (CDATE | CNUMERIC);
	    break;
	  case 'n':
	  case 'N':
	    KeyInfo[KeyCount-1].CompType |= CNUMERIC;
	    break;
	  case 't':
	  case 'T':
	    KeyInfo[KeyCount-1].CompType |= (CTIME | CNUMERIC);
	    break;
	  case 'l':
	  case 'L':
	    KeyInfo[KeyCount-1].CompType &= ~CNUMERIC;
	    break;
	 case 's':
	 case 'S':
	   KeyInfo[KeyCount-1].CompType |= CSIZE; /* Don't set numeric yet */
	   break;
	  default:
	    fprintf(stderr,
		    "Comparison type %s not recognized.\n",optarg);
	    exit(BADOPTIONARG);
	 }
	 break;
       case 'q':		/* Be quiet */
	 VerboseP = FALSE;
	 break;
	 
	 /* Provide information */
	 
       case 'D':		/* List defaults */
	 PrintDefaults(stdout);
	 PrintDefaults(Logfp);
	 exit(SUCCESS);
       case 'F':		/* List flags */
	 PrintFlags(stdout);
	 PrintFlags(Logfp);
	 exit(SUCCESS);
       case 'L':		/* List limits */
	 PrintLimits(stdout);
	 PrintLimits(Logfp);
	 exit(SUCCESS);
       case 'h':
	 PrintUsage();
	 exit(SUCCESS);
       case 'v':		/* This info is always printed so provide flag for compatibility */
	 exit(SUCCESS);
       default:
	 PrintUsage();
	 exit(BADOPTION);
      } /* End switch */
   } /* End getopt while */

   if(KeyCount < 1){
     fprintf(stderr,"%s:no sort keys specified!\n\n",progname);
     exit(OTHERERROR);
   } 
   ProvideDefaultRanks();
   DescribeKeys(Logfp);
   if(ReverseP) fprintf(Logfp,"Order globally inverted.\n");
   if(VerboseP){
	 DescribeKeys(stderr);
	 if(ReverseP) fprintf(stderr,"Order globally inverted.\n");
   }
   
   /* Set field terminators here since defaults are option-dependent */
   
   if(!TerminatorsSetP){
      if(GetRecord == GetBlockSepCharRA){
	 Terminators = create_string(2);
	 Terminators[0] = '\t';
	 Terminators[1] = ' ';
	 Terminators[2] = '\0';
      }
      else{			/* Block */
	 Terminators = create_string(1);
	 Terminators[0] = '\n';
	 Terminators[1] = '\0';
      }
   }

   TerminatorCnt = strlen(Terminators);

   
   /* Set up input and output */
   
   if(ac > optind){
      infile = copy_string(av[optind]);
      infp = OpenFile(infile,"r",progname);
   }
   else{
      infile=copy_string("stdin");
      infp = stdin;
   }
   outfp = stdout;
   
   if(VerboseP) fprintf(stderr,"Reading from %s.\n",infile);
   
   MaxRecords = InitialMaxRecords;
   RecordList = (struct record **) malloc((unsigned) MaxRecords * sizeof(struct record *));
   if(RecordList == (struct record **)0){
      fprintf(stderr,"%s: out of memory - RecordList.\n",progname);
      exit(OUTOFMEMORY);
   }
   
   /* Now we read in one record at a time and extract and store its key */
   
   if(VerboseP)fprintf(stderr,"Records processed:  %9ld",Records);
   while(TRUE){
      InputText = (*GetRecord)(infp,InputText,&RecordLength,&InputTextSize,RecordSeparator);
      if(RecordLength == BUFOVERFLOW){
	fprintf(stderr,"%s: unable to allocate memory for input buffer.\n",progname);
	 exit(OUTOFMEMORY);
      }
      else if(RecordLength == ENDOFINPUT) break;
      else if(RecordLength == 0) continue;
      /* Absorb extra separators - should perhaps be done by input routines */
      else if((RecordLength == 1) && (InputText[0] == RecordSeparator)) continue;
      
      if(Records == MaxRecords){ /* Need more storage? */
	 MaxRecords += InitialMaxRecords/2;
         RecordList = (struct record **) realloc((void *) RecordList,
						 (unsigned) MaxRecords * sizeof(struct record *));
	 if(RecordList == (struct record **)0){
	    fprintf(stderr,"\n%s: out of memory - RecordList.\n",progname);
	    fprintf(stderr,"Storage allocation failed on record %ld.\n",
		    Records+1L);
	    exit(OUTOFMEMORY);
	 }
      }
      
      /* Allocate storage for record structure */
      
      RecordList[Records] = (struct record *) malloc(sizeof(struct record));
      if(RecordList[Records] == (struct record *) 0){
	 fprintf(stderr,"\n%s: out of memory - new record - #%ld.\n",
		 progname,Records+1L);
	 exit(OUTOFMEMORY);
      }
      RecordList[Records]->length = (size_t) RecordLength;
      
      RecordList[Records]->klistptr = (union key *) malloc( (unsigned) KeyCount * sizeof(union key));
      if(RecordList[Records]->klistptr == (union key *) 0){
	 fprintf(stderr,"\n%s: out of memory - keylist of record -#%ld.\n",
		 progname,Records+1L);
	 exit(OUTOFMEMORY);
      }


      /* Store record text */
      
	 RecordList[Records]->text = (char *) malloc( (unsigned) (RecordLength+1) * sizeof(char));
	 if(RecordList[Records]->text == (char *) 0){
	    fprintf(stderr,"\n%s: out of memory - record text -#%ld.\n",
		    progname,Records+1L);
	    exit(OUTOFMEMORY);
	 }
	 strcpy(RecordList[Records]->text,InputText);
	 InputCopy = RecordList[Records]->text;
      
     /* Reallocate TempString if necessary to ensure size exceeds current record */

	 if(TempStringSize <= RecordLength){
	   TempStringSize = RecordLength+1;
	   TempString = (KEYTYPE *) realloc((void *) TempString,
		(unsigned) TempStringSize * sizeof(KEYTYPE));
	 }
	 if(TempString == (KEYTYPE *) 0){
	   fprintf(stderr,"\n%s: Failed to reallocate memory for TempString.\n",progname);
	   exit(OUTOFMEMORY); 
	 }

      /* Now extract and store the sort key */
      
      if(GetKeys(RecordList[Records++],KeyInfo,KeyCount) == ERROR){
	FailedKeyP=1;
	fprintf(Logfp,"Key extraction failed on ill-formed record %ld:\n\n%s\n",Records,
		 InputCopy);
	 /* Just skip the record by decrementing counter. */
	 Records--;
	 /* Free up the storage we already allocated */
	 free(RecordList[Records]->text);
	 free(RecordList[Records]);
      }
      
      /* Report on progress */
      
      if(VerboseP && (Records % REPORTINTERVAL == 0)){
	 fprintf(stderr,"\b\b\b\b\b\b\b\b\b%9ld",Records);
      }	
   } /* End of record-reading loop */
   
   if(VerboseP) fprintf(stderr,"\b\b\b\b\b\b\b\b\b%9ld\n",Records);
   fclose(infp);
   
   if((Records < 2L)&&(VerboseP)){
      fprintf(stderr,"%s: there's no point in sorting fewer than two records.\n",progname);
   }
   
   
   /* Make log entries */
   
   snprintf(PrTempString,PRTEMPSTRLEN,"%ld",Records);
   fprintf(Logfp,"Read %s record%s from %s.\n",DelimitNumber(PrTempString,',',3),
	   (Records > 1L ? "s":""),infile);


   /* Make keys sorted on string length  compare numerically  */
   for(i =0; i < KeyCount; i++) if(KeyInfo[i].CompType & CSIZE) KeyInfo[i].CompType |= CNUMERIC;   

   /* Sort the stored records */
  
   if(VerboseP) fprintf(stderr,"Sorting...");
   if(Records > 1L){
     if( Records > QUICKSORTTHRESHOLD) QuickSort(RecordList,0L, Records-1L);
     else ShellSort(RecordList,Records);
   }
   
   if(VerboseP){
      fprintf(stderr,"\b\b\b\b\b\b\b\b\b\bSort completed. Writing out sorted records.\n");
   }
   
   
   /* Write the sorted records back out */
   
   if(VerboseP) fprintf(stderr,"Records written:    %9ld",0L);
   IntWrite(Records,outfp,ReverseP,RecordSeparator);
   
   /* We're done */

   snprintf(PrTempString,PRTEMPSTRLEN,"%12ld",ComparisonCnt);
   tmpptr = DelimitNumber(PrTempString,',',3);

   fprintf(Logfp,"Comparisons: %s\n",tmpptr);
   if(VerboseP) fprintf(stderr,"Comparisons: %s\n",tmpptr);
   if(FailedKeyP == TRUE){
     fprintf(stderr,
	     "Key extraction failed on one or more records.\nConsult the log for details.\n"); 
   }
   fclose(Logfp);
   exit(SUCCESS);
   
} /* End of main() */


int
SetDateFormat(char *str, struct ymdinfo *ymd)
{
   int i;
   int j;
	
   if(strlen(str) != 5) return(ERROR); 
	
   sprintf(ymd->fmt,"%%d%c%%d%c%%d",str[1],str[3]);
   ymd->y = ymd->m = ymd->d = -1; /* For testing if value has been set */
   for(i = 0, j=0; i < 3; i++,j+=2){
      switch(str[j]){
       case 'y':
	 ymd->y = i;
	 break;
       case 'm':
	 ymd->m = i;
	 break;
       case 'd':
	 ymd->d = i;
	 break;
       default:
	 return(ERROR);
      }
   }
   if(ymd->y < 0 || ymd->m < 0 || ymd->d < 0) return(ERROR);
   else return(SUCCESS);
}

void	
DescribeKeys(FILE *fp)
{
   int i;
   int number;
   char tmp[128];
   extern void DumpExclusions(unsigned char *,int,FILE *);

   if(KeyInfo[0].SelectType == WHOLE){
     fprintf(fp,"Sorting on whole record.\n");
     if(KeyInfo[0].CompType & CINVERSE)fprintf(fp,"Decreasing ");
     else fprintf(fp,"Increasing ");
     if(!(KeyInfo[0].CompType & CNUMERIC)){
       if(KeyInfo[0].ReverseP) fprintf(fp,"reversed ");
       if(KeyInfo[0].CompType & CSIZE) fprintf(fp,"by size\n");
       else fprintf(fp,"lexicographic\n");
       if(KeyInfo[0].MapTableEntries) fprintf(fp," %3d multigraph%s\n",
			      KeyInfo[0].MapTableEntries,
			      (KeyInfo[0].MapTableEntries == 1 ? "":"s"));
       if(KeyInfo[0].ExclusionTable != NULL){
	 DumpExclusions(KeyInfo[0].ExclusionTable,ASCIISIZE,fp);
       }
     }
     else{
       if(KeyInfo[0].CompType & CDATE) fprintf(fp,"date\n");
	 else if(KeyInfo[0].CompType & CTIME) fprintf(fp,"time\n");
	 else fprintf(fp,"numeric\n");
      }
      return;
   }
	
   for(i = 0;i < KeyCount; i++){
      fprintf(fp,"Key %d ",i+1);
      fprintf(fp,"%-11.11s ",(KeyInfo[i].OptionalP ? "optional":"obligatory"));
      if(KeyInfo[i].SelectType == NUMBERED){
	number = 1+KeyInfo[i].Number;
	if(number > 0){
	  if(number == 1) sprintf(tmp,"first field");
	  else if(number == 2) sprintf(tmp,"second field");
	  else if(number == 3) sprintf(tmp,"third field");
	  else if(number == 4) sprintf(tmp,"fourth field");
	  else if(number == 5) sprintf(tmp,"fifth field");
	  else if(number % 100 == 11) sprintf(tmp,"%dth field",number);
	  else if(number % 100 == 12) sprintf(tmp,"%dth field",number);
	  else if(number %  10 == 1) sprintf(tmp,"%dst field",number);
	  else if(number %  10 == 2) sprintf(tmp,"%dnd field",number);
	  else sprintf(tmp,"%dth field",number);
	}
	else{
	  number *= -1;
	  if(number == 1) sprintf(tmp,"last field");
	  else if(number == 2) sprintf(tmp,"penultimate field");
	  else if(number == 3) sprintf(tmp,"antepenultimate field");
	  else if(number == 4) sprintf(tmp,"fourth field from end");
	  else if(number == 5) sprintf(tmp,"fifth field from end");
	  else if(number % 100 == 11) sprintf(tmp,"%dth field from end",number);
	  else if(number % 100 == 12) sprintf(tmp,"%dth field from end",number);
	  else if(number %  10 == 1) sprintf(tmp,"%dst field from end",number);
	  else if(number %  10 == 2) sprintf(tmp,"%dnd field from end",number);
	  else sprintf(tmp,"%dth field from end",number);
	}
      }
      else sprintf(tmp,"tag %s",KeyInfo[i].Tag);
      fprintf(fp,"%-26.26s ",tmp);
      if(KeyInfo[i].CompType & CINVERSE)fprintf(fp,"Decreasing ");
      else fprintf(fp,"Increasing ");
      if(!(KeyInfo[i].CompType & CNUMERIC)){
	 if(KeyInfo[i].ReverseP) fprintf(fp,"reversed ");
	 if(KeyInfo[i].CompType & CSIZE) fprintf(fp,"by size\n");
	 else fprintf(fp,"lexicographic\n");
	 if(KeyInfo[i].MapTableEntries) fprintf(fp," %3d multigraph%s\n",
						KeyInfo[i].MapTableEntries,
						(KeyInfo[i].MapTableEntries == 1 ? "":"s"));
	 if(KeyInfo[i].ExclusionTable != NULL){
	    DumpExclusions(KeyInfo[i].ExclusionTable,ASCIISIZE,fp);
	 }
      }
      else{
	 if(KeyInfo[i].CompType & CDATE) fprintf(fp,"date\n");
	 else if(KeyInfo[i].CompType & CTIME) fprintf(fp,"time\n");
	 else fprintf(fp,"numeric\n");
      }
   }
}

double *
AllocateDouble()
{
   double *new;
	
   new = (double *) malloc(sizeof(double));
   if(new == NULL){
      fprintf(stderr,"%s: out of memory - numeric key\n",progname);
      exit(OUTOFMEMORY);
   }
   return(new);
}


/*
 * Extract the keys from the record, map them into internal form
 * and store them with the record.
 */	

int
GetKeys(struct record *recptr, struct keyinfo *info,int keys)
{
   char *Next;
   char *FStart;
   char **FieldPtrs;
   int i, j;
   int FieldCnt;
   int MaxFields;
   int KeyLength;
   int num;			/* Dereferenced field number */
   int index;			/* Index into array of fields */
   double TempNumKey;
   short FoundP;
   short KeyMissingP;
   
   void KeyCopy(KEYTYPE *, KEYTYPE *, int);
   void RevKeyCopy(KEYTYPE *, KEYTYPE *, int);
   int MapKey(char *, KEYTYPE *,struct mapentry **,int);
	
   extern char *strtok();
   extern void *malloc(size_t);
   extern int Exclude(KEYTYPE *,int,unsigned char *);
   extern int GetTimeKey(char *, double *);
   extern int GetDateKey(char *, double *,struct ymdinfo *);
	
   /*
    * If the whole record is to be used as key, skip the parse and
    * treat the whole record as the first key.
    */

   if(info[0].SelectType == WHOLE){
      MaxFields = 1;
      FieldPtrs = (char **) malloc((unsigned) MaxFields * sizeof(char *));
      if(FieldPtrs == (char **)0){
	 fprintf(stderr,"%s: out of memory - FieldPtrs.\n",progname);
	 exit(OUTOFMEMORY);
      }
      FieldCnt = 1;
      FieldPtrs[0] = &(InputText[0]);
      if(info[0].CompType & CNUMERIC){
	 if(info[0].CompType & CTIME){
	    if(GetTimeKey(FieldPtrs[0],&TempNumKey) == ERROR){
	       fprintf(stderr,"\n%s is not a well-formed time.\n",FieldPtrs[0]);
	       fprintf(Logfp,"\n%s is not a well-formed time.\n",FieldPtrs[0]);
	       return(ERROR);
	    }
	 }
	 else if(info[0].CompType & CDATE){
	    if(GetDateKey(FieldPtrs[0],&TempNumKey,&(info[0].ymd)) == ERROR){
	       fprintf(stderr,"\n%s is not a well-formed date.\n",FieldPtrs[0]);
	       fprintf(Logfp,"\n%s is not a well-formed date.\n",FieldPtrs[0]);
	       return(ERROR);
	    }
	 }
	 else TempNumKey = atof(FieldPtrs[0]);
	 recptr->klistptr[0].n = AllocateDouble();
	 *(recptr->klistptr[0].n) = TempNumKey;
      }
      else{			/* Lexicographic key */
	 KeyLength = MapKey(FieldPtrs[0],
			    TempString,
			    info[0].MapTable,
			    info[0].MapTableEntries);
			
	 /* Handle exclusions */
			
	 if(info[0].ExclusionTable != NULL){
	    KeyLength = Exclude(TempString,KeyLength,info[0].ExclusionTable);
	    if(KeyLength == 0){
	      fprintf(stderr,"After exclusions record is empty.\n");
	      return(ERROR);
	    }
	 }

	 /* Store the key */

	if(info[0].CompType & CSIZE){
	    recptr->klistptr[0].n = AllocateDouble();
	    *(recptr->klistptr[0].n) = (double) KeyLength;
	}
	else{
	 recptr->klistptr[0].t = (KEYTYPE *) malloc(sizeof(KEYTYPE) * (KeyLength +1));
	 if(recptr->klistptr[0].t == (KEYTYPE *) 0){
	    fprintf(stderr,"\n%s: out of memory - new key.\n",
		    progname);
	    exit(OUTOFMEMORY);
	 }
	 if(info[0].ReverseP) RevKeyCopy(recptr->klistptr[0].t,TempString,KeyLength+1);
	 else KeyCopy(recptr->klistptr[0].t,TempString,KeyLength+1);
	}
      }
      free((void *) FieldPtrs);      
      return (SUCCESS);
   }
   else{
		
      /*
       * Parse the record into fields, leaving the number of fields
       * found in Fieldcnt and pointers to the beginnings in FieldPtrs.
       */	    

      /* Allocate minimum number of field pointers here */
      MaxFields = InitialMaxFields;
      FieldPtrs = (char **) malloc((unsigned) MaxFields * sizeof(char *));
      if(FieldPtrs == (char **)0){
	 fprintf(stderr,"%s: out of memory - FieldPtrs.\n",progname);
	 exit(OUTOFMEMORY);
      }

      FieldCnt = 0;
      Next = &(InputText[0]);
      while((FStart = strtok(Next,Terminators)) != NULL){
	 if(FieldCnt == MaxFields){/* Reallocate pointers here */
	    MaxFields += InitialMaxFields/2;
	    FieldPtrs = (char **) realloc((void *) FieldPtrs,
			  (unsigned) MaxFields * sizeof(char *));
	    if(FieldPtrs == (char **)0){
	       fprintf(stderr,"\n%s: out of memory - FieldPtrs.\n",progname);
	       exit(OUTOFMEMORY);
	    }
	 }
	 FieldPtrs[FieldCnt++] = FStart;
	 Next = NULL;
      }

      /*
       * For each key, find the required field and map.
       */	    

      for(i = 0; i < keys; i++){
	 FoundP = FALSE;
	 KeyMissingP = FALSE;
	 if(info[i].SelectType == NUMBERED){
	    num = info[i].Number;
	    if(num < 0) index = FieldCnt + num +1;
	    else index = num;
	    if((index >= 0) && (index < FieldCnt)){
	       if(info[i].CompType & CNUMERIC){
		  if(info[0].CompType & CTIME){
		     if(GetTimeKey(FieldPtrs[index],&TempNumKey) == ERROR){
			fprintf(stderr,"\n%s is not a well-formed time\n",FieldPtrs[index]);
			fprintf(Logfp,"\n%s is not a well-formed time\n",FieldPtrs[index]);
			return(ERROR);
		     }
		  }
		  else if(info[i].CompType & CDATE){
		     if(GetDateKey(FieldPtrs[index],
				   &TempNumKey,&(info[i].ymd)) == ERROR){
			fprintf(stderr,"\n%s is not a well-formed date.\n",FieldPtrs[index]);
			fprintf(Logfp,"\n%s is not a well-formed date.\n",FieldPtrs[index]);
			return(ERROR);
		     }
		  }
		  else TempNumKey = atof(FieldPtrs[index]);
	       }
	       else{		/* Lexicographic key */
		  KeyLength = MapKey(FieldPtrs[index],
				     TempString,
				     info[i].MapTable,
				     info[i].MapTableEntries);
	       }
	    }
	    else{		/* Key field not present */
	       if(info[i].OptionalP == FALSE){ /* If not optional, gripe */
		  fprintf(stderr, "\n%s: record does not have %d fields.\n", progname,num+1);
		  fprintf(Logfp, "\n%s: record does not have %d fields.\n", progname,num+1);
		  return(ERROR);
	       }
	       else KeyMissingP=TRUE;
	    }
	 }
	 else{			/* SelectType = Tagged */
	    for(j = 0;j < FieldCnt; j++){
	       if(regexec(info[i].cregexp,FieldPtrs[j])){
		  FoundP = TRUE;
		  if(info[i].CompType & CNUMERIC){
		     if(info[0].CompType & CTIME){
			if(GetTimeKey(info[i].cregexp->endp[1],
				      &TempNumKey) == ERROR){
			   fprintf(stderr,"\n%s is not a well-formed time\n",
				   info[i].cregexp->endp[1]);
			   fprintf(Logfp,"\n%s is not a well-formed time\n",
				   info[i].cregexp->endp[1]);
			   return(ERROR);
			}
		     }
		     else if(info[i].CompType & CDATE){
			if(GetDateKey(info[i].cregexp->endp[1],
				      &TempNumKey,&(info[i].ymd)) == ERROR){
			   fprintf(stderr,"\n%s is not a well-formed date.\n",
				   info[i].cregexp->endp[1]);
			   fprintf(Logfp,"\n%s is not a well-formed date.\n",
				   info[i].cregexp->endp[1]);
			   return(ERROR);
			}
		     }
		     else TempNumKey = atof(info[i].cregexp->endp[1]);
		  }
		  else{		/* Lexicograhic key */
		     KeyLength = MapKey(info[i].cregexp->endp[1],
					TempString,
					info[i].MapTable,
					info[i].MapTableEntries);
		  }
		  break;					 
	       }
	    }
	    if(FoundP == FALSE){
	       if(info[i].OptionalP == FALSE){ /* If not optional, gripe */
		  fprintf(Logfp,
			  "\nRecord does not have a field with tag matching \"%s\" on which to sort.\n",info[i].Tag);
		  return(ERROR);
	       }
	       else KeyMissingP=TRUE;
	    }
	 }

	 /* Handle exclusions */

	 if( (info[i].ExclusionTable != NULL) && !KeyMissingP){
	    KeyLength = Exclude(TempString,KeyLength,info[i].ExclusionTable);
	    if(KeyLength == 0){
	      if(!info[i].OptionalP){
		fprintf(stderr,"After exclusions non-optional field is empty.\n");
		return(ERROR);
	      }
              KeyMissingP=TRUE;
	    }
	 }

	 /* Store the key */

	 if(info[i].CompType & CNUMERIC){
	   if(!KeyMissingP){
	    recptr->klistptr[i].n = AllocateDouble();
	    *(recptr->klistptr[i].n) = TempNumKey;
	   }
	   else recptr->klistptr[i].n = NULL;
	 }
	 else{			/* Lexicographic key */
	   if(!KeyMissingP){
		if(info[i].CompType & CSIZE){
		    recptr->klistptr[i].n = AllocateDouble();
		    *(recptr->klistptr[i].n) = (double) KeyLength;
		 }
		else{
		    recptr->klistptr[i].t = (KEYTYPE *) malloc(sizeof(KEYTYPE) * (KeyLength +1));
	            if(recptr->klistptr[i].t == (KEYTYPE *) 0){
		         fprintf(stderr,"\n%s: out of memory - new key.\n", progname);
		         exit(OUTOFMEMORY);
		    }
		 if(info[i].ReverseP) RevKeyCopy(recptr->klistptr[i].t,TempString,KeyLength+1);
	         else KeyCopy(recptr->klistptr[i].t,TempString,KeyLength+1);
		}
	   }
	   else{
	     if(info[i].CompType & CSIZE) recptr->klistptr[i].n = NULL;  
	     else recptr->klistptr[i].t = NULL;   
	   }
	 }
      }	/* End of loop over fields */
      free((void *) FieldPtrs);
      return (SUCCESS);
   }
}

int
GetSortOrder(char *file,struct keyinfo *ki)
{
   FILE *fp;
   
   extern FILE *OpenFile(char *, char *, char *);
   void SortMapTable(struct keyinfo *);
   void CreateRankTable(struct keyinfo *);
	
   fp = OpenFile(file,"r",progname);
   CreateRankTable(ki);
   if(ConstructSortOrder(fp,ki) == ERROR)return(ERROR);
   fclose(fp);
   SortMapTable(ki);
   return(SUCCESS);
}

/*
 * Initialize all characters into ASCII order but so that,
 * if left unset, they will be ranked after all characters whose
 * order is set.	 
 */	 

void
CreateRankTable(struct keyinfo *ki)
{
   int i;
	
   extern void *malloc(size_t);
	
   ki->RankTable = (int *) malloc(sizeof(int) * RANKTABLESIZE);
   if(ki->RankTable == (int *) 0){
      fprintf(stderr,"Out of memory.\n");
      exit(OUTOFMEMORY);
   }
	
   /*
    * Code 0 must rank first so that a string will sort before
    *	a string of which it is a prefix.
    */
   ki->RankTable[0] = 0;
	
   /* This yields machine collating order but after all user-specified codes */
   for(i=1;i<RANKTABLESIZE;i++)ki->RankTable[i] = RANKTABLESIZE + i +1;
}

void
InitializeKeyInfo(void)
{
   int i;
	
   for(i = 0;i < MAXKEYS; i++){
      KeyInfo[i].CompType = (unsigned char) (0 & ~CNUMERIC & ~CINVERSE);
      KeyInfo[i].OptionalP = FALSE;
      KeyInfo[i].MissingKeyComparison = COMPARE_EQUAL;
      KeyInfo[i].ReverseP = FALSE;
      KeyInfo[i].MapTable = NULL;
      KeyInfo[i].MapTableEntries = 0;
      KeyInfo[i].NextCode = ASCIISIZE;
      KeyInfo[i].RankTable = (int *) 0;
      KeyInfo[i].ExclusionTable = (unsigned char *) 0;
      KeyInfo[i].ExclusionSpec=0;
      KeyInfo[i].ymd.y = 0;
      KeyInfo[i].ymd.m = 1;
      KeyInfo[i].ymd.d = 2;
      sprintf(KeyInfo[i].ymd.fmt,"%%d/%%d/%%d");
   }
}

/* Write the internally stored records out in sorted order */

void
IntWrite(long Records, FILE *fp,int ReverseP,char sep)
{
   long i;
   void PutIntRecord(struct record *,FILE *, char);

   for(i=1L; i <= Records; i++){
      PutIntRecord(RecordList[(ReverseP ? Records-i : i-1)],fp,sep);
      if(VerboseP && (i % INTWREPORTINTERVAL == 0)){
	 fprintf(stderr,"\b\b\b\b\b\b\b\b\b%9ld",i);
      }
   }
   if(VerboseP) fprintf(stderr,"\b\b\b\b\b\b\b\b\b%9ld\n",i-1);
}

/* Write out a single stored record */

void
PutIntRecord(struct record *ptr,FILE *fp, char sep)
{
   fputs(ptr->text,fp);
   fputc(sep,fp);
}

int
ConstructSortOrder(FILE *sfp, struct keyinfo *ki)
{
   int CurrentRank = 0;
   char *Field;
   char *str;
   char *InputLine;
   int EndOfFileP;
   long LineLength;
   unsigned int Index;
   
   extern char * GetArbChunk(FILE *, char, long *, int *);
   extern int EvalEscapes(char *);
   KEYTYPE AddStringToMapTable(char *,struct keyinfo *);

   extern char *strtok();
	
   while(TRUE){
      InputLine=GetArbChunk(sfp,(char)'\n',&LineLength,&EndOfFileP);
      if(InputLine == (char *) 0){
	 fprintf(stderr,"Storage allocation failure on sort order file.\n");
	 fprintf(stderr,"Failed on line of length %ld.\n",LineLength);
	 return ERROR;
      }
      else if(EndOfFileP == 1) break;
      ++CurrentRank;
      str = &(InputLine[0]);
      while((Field = strtok(str," \t\0")) != NULL){
	 str = NULL;
	 if(EvalEscapes(Field) == ERROR){
	    fprintf(stderr,"EvalEscapes: out of memory.\n");
	    exit(OUTOFMEMORY);
	 }
	 if(strlen(Field) > 1){
	    ki->RankTable[Index = (0x0000FFFF & ((unsigned int) AddStringToMapTable(Field,ki)))] = CurrentRank;
	 }
	 else ki->RankTable[Index = (0x000000FF & ((unsigned int)(Field[0])))] = CurrentRank;
	 /*	 fprintf(stderr,"%8s\t rank of %6d = %6d\n",Field,Index,CurrentRank); */
      }
   }
   free((void *)InputLine);
   return(SUCCESS);
}


struct mapentry **
AllocateMapTable()
{
   struct mapentry **new;
	
   new = malloc(sizeof(struct mapentry *) * MAXMULTIGRAPHS);
   if(new == NULL){
      fprintf(stderr,"%s: out of memory - map table\n",progname);
      exit(OUTOFMEMORY);
   }
   return(new);
}

/*
 * Add a string to the mapping table and return the code allocated to it.
 */

KEYTYPE
AddStringToMapTable(char *s, struct keyinfo *ki)
{
   struct mapentry *new;
   
   extern void *malloc(size_t);
   
   if(ki->MapTable == NULL) ki->MapTable = AllocateMapTable();

   if(ki->NextCode == RANKTABLESIZE){
      fprintf(stderr,"%s:Too many multigraphs defined.\n",progname);
      exit(LIMITEXCEEDED);
   }
   
   /* Allocate a new map table entry */
   
   new = (struct mapentry *) malloc(sizeof(struct mapentry));
   if(new == (struct mapentry *) 0){
      fprintf(stderr,"%s: out of memory - map table entry\n",progname);
      exit(OUTOFMEMORY);
   }
   
   new->len = strlen(s);
   
   new->str = (char *) malloc((new->len + 1) * sizeof(char));
   if(new->str == (char *) 0){
      fprintf(stderr,"%s:out of memory - map table string\n",progname);
      exit(OUTOFMEMORY);
   }
   strcpy(new->str,s);
   new->code = ki->NextCode;

   ki->MapTable[ki->MapTableEntries]=new;
   ki->MapTableEntries+=1;
   
   return((ki->NextCode)++);
}


/* Sort the map table so that sequence length is decreasing */

void
SortMapTable(struct keyinfo *ki)
{
   int gap;
   int i;
   int j;
   
   struct mapentry *temp;
   
   for(gap = ki->MapTableEntries/2; gap > 0; gap /= 2){
      for(i = gap; i < ki->MapTableEntries; i++){
	 for(j = i-gap; j>= 0; j -= gap){
	    if(ki->MapTable[j]->len >= ki->MapTable[j+gap]->len) break;
	    else{
	       temp = ki->MapTable[j];
	       ki->MapTable[j] = ki->MapTable[j+gap];
	       ki->MapTable[j+gap] = temp;
	    }
	 }
      }
   }
}

int
GetExclusions(char *file,struct keyinfo *ki)
{
  FILE *fp;
  char cspec[6];
  char pos[5];
  int i;
  int fields;
  unsigned char flag;
  int poslen;
  int ExCount;
   
  extern FILE *OpenFile(char *, char *, char *);
  unsigned char *CreateExclusionTable(void);
   
  ExCount = 0;
  fp = OpenFile(file,"r",progname);
  ki->ExclusionTable = CreateExclusionTable();
  while( (fields = fscanf(fp,"%4s%3s",cspec,pos)) != EOF){
    if(fields != 2){
      fprintf(stderr,"Format error in exclusion file %s.\n",file);
      exit(OTHERERROR);
    }
    if(EvalEscapes(cspec) == ERROR){
      fprintf(stderr,"EvalEscapes: out of memory.\n");
      exit(OUTOFMEMORY);
    }
    flag = 0;
    poslen = strlen(pos);
    for(i = 0; i < poslen; i++){
      switch(pos[i]){
      case 'F':
      case 'f':
	flag |= EX_FINAL;
	break;
      case 'I':
      case 'i':
	flag |= EX_INITIAL;
	break;
      case 'M':
      case 'm':
	flag |= EX_MEDIAL;
	break;
      default:
	fprintf(stderr,
		"Invalid exclusion position specifier %c in file %s\n",
		pos[i],file);
	exit(OTHERERROR);
      }
    }
    ki->ExclusionTable[cspec[0]] = flag;
    ++ExCount;
  }
  fclose(fp);
  if(ExCount < 1){
    free((void *) ki->ExclusionTable);
    ki->ExclusionTable = NULL;
  }
  return(SUCCESS);
}

unsigned char *
CreateExclusionTable(void)
{
   unsigned char *new;
   int i;

   extern void *malloc(size_t);
	
   new = (unsigned char *) malloc(sizeof(unsigned char) * ASCIISIZE);
   if(new == (unsigned char *) 0){
      fprintf(stderr,"Out of memory.\n");
      exit(OUTOFMEMORY);
   }
   for(i = 0; i < ASCIISIZE;i++) new[i] = (unsigned char) 0;
   return(new);
}

int
GetExclusionsString(char *string,struct keyinfo *ki)
{
   unsigned char *CreateExclusionTable(void);
   int len;
   int ExCount;
   char *exstr;
   int i; 

   extern char *copy_string(char *);

   ExCount = 0;

   ki->ExclusionTable = CreateExclusionTable();
   exstr = copy_string(string);
   if(EvalEscapes(exstr) == ERROR){
     fprintf(stderr,"EvalEscapes: out of memory.\n");
     exit(OUTOFMEMORY);
   }
   len= strlen(exstr);
   for(i=0;i<len;i++){
     ki->ExclusionTable[exstr[i]] = (EX_FINAL | EX_INITIAL | EX_MEDIAL);
     ++ExCount;
   }
}


/*
 * Map key string characters and sequences to internal sort codes.
 * This presupposes that the map table has been sorted by SortMapTable()
 * into decreasing order of length.
 *	 
 *	Start at the beginning of the input string.
 * While there are still characters in the input string
 * 	Set the index into the MapTable to zero
 *		Work through MapTable, looking for prefixes of the input string.
 * 	If we find a prefix
 *			copy the corresponding code to the output
 *			advance the input pointer by the length of the prefix
 *			try again
 *		If we don't find a prefix
 *			copy the current character to the output
 *			advance the input pointer one character
 *			try again
 * Else we're done mapping this string
 */	 

int
MapKey(char *in,KEYTYPE *out,struct mapentry **table,int entries)
{
   int mte;
   int j = 0;
   int FoundP;
   
   if(entries == 0){		/* Don't bother with fancy mapping */
      while(*in != '\0') out[j++] = 0x00FF & (KEYTYPE) *in++;
      out[j]= (KEYTYPE) '\0';
      return(j);
   }
   
   FoundP = FALSE;
   while(*in != '\0'){
      mte = 0;
      FoundP = FALSE;
      while(mte < entries){
	 if(strncmp(in,table[mte]->str,table[mte]->len) == 0){
	    FoundP = TRUE;
	    out[j++] = table[mte]->code;
	    in+=table[mte]->len;
	    break;
	 }
	 ++mte;
      }
      if(!FoundP) out[j++] = 0x00FF & (KEYTYPE) *in++;
   }
   out[j]= (KEYTYPE) '\0';
   return(j);
}


/*
 *  This is the recursive routine in K&R with swap in-lined as a macro
 *  and use of a straight insertion sort for small sub-partitions.
 */	 


#define SWAP(a,b,c) temp = a[b];a[b] = a[c];a[c] = temp;

void
QuickSort(struct record **rlist, long left, long right)
{
   long i;
   long last;
   long j;
   long elts;
   struct record *temp;
	
   int Compare(struct record *,struct record *);

   elts = right - left + 1L;
   if(elts < 2L) return;
	
   if(elts < INSERTIONSORTTHRESHOLD){ /* Do a straight insertion sort */
      for(j=1L;j < elts;j++){
	 temp = rlist[j+left];
	 i = j -1L;
	 while( (i >= 0) && (Compare(rlist[i+left],temp) > 0)){
	    rlist[i+1L+left] = rlist[i+left];
	    --i;
	 }
	 rlist[i+1L+left] = temp;
      }
      return;
   }
	
   /* If we get here it is worth continuing with quicksort. */
	
   j = (left + right)/2L;
   SWAP(rlist,left,j)
	
   last = left;
   for(i = left+1; i <= right; i++){
      if(Compare(rlist[i],rlist[left]) < 0){
	 ++last;
	 SWAP(rlist,last,i);
      }
   }
   SWAP(rlist,left,last);
   QuickSort(rlist,left,last-1L);
   QuickSort(rlist,last+1L,right);
}

void
ShellSort(struct record **rlist,long cnt)
{
   long gap;
   long i;
   long j;
   
   struct record *temp;
	
   int Compare(struct record *,struct record *);
   
   for(gap = cnt/2; gap > 0; gap /= 2){
      for(i = gap; i < cnt; i++){
	 for(j = i-gap; j>= 0; j -= gap){
            if(Compare(rlist[j],rlist[j+gap]) <= 0) break;
	    else{
	       temp = rlist[j];
	       rlist[j] = rlist[j+gap];
	       rlist[j+gap] = temp;
	    }
	 }
      }
   }
   return;
}

inline int
sign(double x)
{
   if(x > 0.0) return (1);
   else if (x < 0.0) return (-1);
   else return (0);
}

/*
 * This is the subroutine that does the actual comparison of two records.
 * It iterates over the keys, returning as soon as it finds a key on
 * which the two records compare differently. For each key it first
 * tests to see if the records have values for that key. If they do,
 * it does the actual comparison, the details depending on whether the
 * comparison is numeric (including dates, times, and sizes), or lexicographic.
 * If both records lack the key, it returns 0. If one record lacks the
 * key and the other does not, it returns a value depending on how
 * MissingKeyComparison is set for this key.
 *
 * Whether a key exists for a particular record is determined by
 * testing whether the pointer is null. It is therefore important
 * that key pointers be set to null during key extraction if a key is missing. 
 */

int
Compare(struct record *a,struct record *b)
{
   int i;
   int rval;
   KEYTYPE *ap;
   KEYTYPE *bp;
   int *rt;
   unsigned short APresent;
   unsigned short BPresent;

   ComparisonCnt++;

   for(i = 0; i < KeyCount; i++){
     /* First set flags for presence of key in each record */
     APresent = BPresent = FALSE;
     if(KeyInfo[i].CompType & CNUMERIC){
       if(a->klistptr[i].n != NULL) APresent=TRUE;
       if(b->klistptr[i].n != NULL) BPresent=TRUE;
     }
     else{
       ap = a->klistptr[i].t;
       if(ap != NULL) APresent=TRUE;
       bp = b->klistptr[i].t;
       if(bp != NULL) BPresent=TRUE;
     }

     /* If both records possess key proceed with empirical comparison*/
     if(APresent && BPresent){
       if(KeyInfo[i].CompType & CNUMERIC){
	 if(KeyInfo[i].CompType & CINVERSE){
	   rval = sign(*(b->klistptr[i].n) - *(a->klistptr[i].n));
	 }
	 else rval = sign(*(a->klistptr[i].n) - *(b->klistptr[i].n));
       }
       else{
	 rt = KeyInfo[i].RankTable;
	 rval = 1;		/* Use as flag to see if we reached end of string */ 
	 
	 for( ; rt[*ap] == rt[*bp]; ap++, bp++){
	   if(*ap == (KEYTYPE) '\0'){rval=0;break;} 
	 }
	 if(rval != 0){
	   if(KeyInfo[i].CompType & CINVERSE) rval = rt[*bp] - rt[*ap];
	   else rval = rt[*ap] - rt[*bp];
	 }
       }
     }
     /* If we get here at least one record lacks this key */
     else{
       if (APresent && !BPresent) rval = 1 - KeyInfo[i].MissingKeyComparison;
       else{
	 if (!APresent && BPresent) rval = KeyInfo[i].MissingKeyComparison - 1;
	 else{
	   if (!APresent && !BPresent)rval = 0;
	 }
       }
     }
     if(rval != 0) return(rval); /* If records are same on this key, go on to next key */
   } /* End of loop over keys */
   return(0);			/* If we get here, records tie on all keys */
}

void
ProvideDefaultRanks(void)
{
   int i;
	
   for(i = 0; i < KeyCount;i++){
      if((!(KeyInfo[i].CompType & CNUMERIC)) && (KeyInfo[i].RankTable == NULL)){
	 CreateRankTable(&(KeyInfo[i]));
      }
   }
}

void
Copyright(FILE *fp)
{
  fprintf(fp,"Copyright 1993-2004 William J. Poser (billposer@alum.mit.edu)\n");
  fprintf(fp,"Released under the terms of the GNU General Public License.\n");
}

